#include <iostream>
#include <cstdlib>
#include <cctype>
#include <csignal>
#include <fstream>

#include "zero.hh"
#include "kertoimet.hh"
#include "vakiot.hh"
#include "yksikot.hh"
#include "yksikko.hh"
#include "summa.hh"
#include "tulo.hh"

extern void Dump(elemp res);

Kertoimet         Kert;
Yksikot           Yks;
Yksikkomuunnokset Ymu(Yks);
Vakiot            Vak(Yks);

static bool VirallinenTyyli = false;
static bool Optimization    = true;
static bool dumpdebug       = false;
static bool Grammahallinta  = true;
static bool SIhallinta      = true;
        
class BadInput
{
    string desc;
public:
    BadInput() { }
    BadInput(const char *s) : desc(s) { }
    const string& value() const { return desc; }
};

static void LoadVariables(void);
static void SaveVariable(const string &var, const string &s);
static class Parser
{
public:
    static void delblank(string &s)
    {
        while(s.size() && isspace(s[0]))s.erase(0, 1);
    }
    static merknum fetchnum(string &str) /* const is not valid for static */
    {
        char *ep;
        const char *s = str.c_str();
        const char *const begin = s;
        
        real val = strtold(s, &ep);
        
        // cout.precision(35);
        // cout << "value read: " << val << endl;
        
        /* Got the value, next analyze the precision */
        unsigned precis = 0;
        
        const char *q = s;
        if(*q == '+' || *q == '-')++q;
        bool pilkku = false;
        unsigned nollia=0;
        for(; q < ep; ++q)
        {
            if(*q == '0') { if(pilkku&&precis)++precis;else ++nollia; continue; }
            if(*q >= '1' && *q <= '9')
            {
                if(!pilkku)precis += nollia;
                ++precis;
                nollia = 0;
                continue;
            }
            /* Do not analyze the exponent */
            if(*q == 'E' || *q == 'e')break;
            if(*q == '.')
            {
                if(precis)precis += nollia;
                pilkku = true;
                nollia = 0;
            }
        }
        if(VirallinenTyyli || pilkku || precis==0)
        {
            // Virallisessa 1000:n tarkkuus on 4
            // Joka tapauksessa 0.100:n tarkkuus on 3
            precis += nollia;
        }
        
        str.erase(0, ep-begin);
        delblank(str);
        if(str.size() && str[0]=='')
        {
            str.erase(0, 1);
            if(val > 0)
                val += 0.5;
            else
                val -= 0.5;
            delblank(str);
        }
        if(str.size() && str[0]=='#')
        {
            str.erase(0, 1);
            precis = TARKKUUS;
            delblank(str);
        }
        return merknum(val, precis);
    }
    static tulo *ParseTulo(string &s, merknum &kerroin)
    {
        tulo *tulotmp = new tulo;
        
        bool jako = false;
        
        for(;;)
        {
            delblank(s);
            if(!s.size())break;
            
            unsigned a, b;
            int k = 0;
            
            elemp res;
            
            merknum ker = fetchnum(s);
            if(!ker.valid())ker = 1;
            
            //out << "ker[" << ker.value() << "," << ker.precision() << ']';
            
            a = Kert.Test(s);
            if(a && (b = Yks.Test(s, a)) && Yks.SI(Yks.Value(s, a))) // Ei SI-kertoimettomia.
            {
                k = Kert.Value(s);
                /* k:n "yksikk": 10^ */
            YksLabel:
                char y = Yks.Value(s, a);
                s.erase(0, a+b);
                
                ker *= pow10l(k);
                res = new yksikko(y);
            }
            else if((a = Vak.Test(s)))
            {
                //cout << "(max precision with " << s << ")";
                res = &*Vak.Value(s);
                s.erase(0, a);
                
                ker.maxprecision();
            }
            else /* a=0 */
            {
                b = Yks.Test(s);
                if(b)
                {
                    /* k = 0; */
                    goto YksLabel;
                }
                /* a=0, b=0 */
                res = NULL;
                
                if(s.size() && s[0] == '(')
                {
                    s.erase(0, 1);
                    res = ParseSumma(s);
                    delblank(s);
                    if(s[0] != ')')throw BadInput("Expected: ')'");
                    s.erase(0, 1);
                }
            }
            
            real pot = 1;
            delblank(s);
            if(s.size())
            {
                if(s[0]=='^')
                {
                    s.erase(0, 1);
                    delblank(s);
                    pot = fetchnum(s).value();
                }
            }
            while(s.size())
            {
                if(s[0]!='' && s[0]!='')break;
                if(pot==1)pot=0;
                pot=pot*10;
                pot += (s[0]=='') ? 2 : 3;
                s.erase(0, 1);
            }
            
            if(res)
            {
                tulotmp->insert(res, jako ? -pot : pot);
            }
            
            if(jako)
                kerroin /= powl(ker, pot);
            else
                kerroin *= powl(ker, pot);
            kerroin.combine(ker);
            
            jako = false;
            if(s.size() && s[0]=='*')
            {
                s.erase(0, 1);
            }
            else if(s.size() && s[0]=='/')
            {
                jako = true;
                s.erase(0, 1);
            }
            else if(!res || (s.size() && (s[0]=='-' || s[0]=='+')))break;
        }
        return tulotmp;
    }
    static elemp ParseSumma(string &s)
    {
        summa *summatmp = new summa;
        bool miinus = false;
        for(;;)
        {
            delblank(s);
            if(!s.size())break;
            
            merknum kerroin = 1;
            tulo *p = ParseTulo(s, kerroin);
            if(!p)break;
            
            if(p->size())
                summatmp->insert(p, miinus ? -kerroin : kerroin);
            else if(miinus)
                summatmp->lisays -= kerroin;
            else
                summatmp->lisays += kerroin;
            
            delblank(s);
            if(s.size() && s[0]=='+')
            {
                miinus = false;
                s.erase(0, 1);
            }
            else if(s.size() && s[0]=='-')
            {
                miinus = true;
                s.erase(0, 1);
            }
            else break;
        }
        return summatmp;
    }
    
    static bool isidchar(char c)
    {
        return isalnum(c) || c=='.' || c=='$' || c=='[' || c==']';
    }
    static string setvarname(string &s, bool checker=true);
    static void Parse(string s, bool initvar=false)
    {
        bool vihre = false;
        
        delblank(s);
        string resultvar;
        
        elemp result;
        try
        {
            resultvar = setvarname(s);
            if(!resultvar.size())resultvar = "result";
            if(!initvar)LoadVariables();
            
            result = ParseSumma(s);
        }
        catch(BadInput &t)
        {
            cout << "ce: " << t.value() << endl;
            vihre = true;
            result = NULL;
        }
        if(s.size())
        {
            cout << "ce: Parse error before \"" << s << "\"";
            if(!ispunct(s[0]))cout << " (unrecognized identifier)";
            cout << endl;
            vihre = true;
        }
        if(vihre)return;
        
        if(Optimization)result = result->Expand();
        
        if(initvar)
        {
            // cout << resultvar << " = " << DumpResult(result) << endl;

            autoeqptr<summa> tmp = new summa;
            tmp->insert(&*result, 1);
            Vak[resultvar] = tmp;
        }
        else
        {
            string res = DumpResult(result);
            cout << resultvar << " = " << res << endl;
        
            SaveVariable(resultvar, res);
        }
    }

    static string DumpMerkNum(merknum e);
    static int SuggestKerroin(const merknum &e, bool SentitYms = false);
    static string DumpResult(elemp result, int suggkerr = 0);
} p;

static void LoadVariables(void)
{
    const char *home = getenv("HOME");
    string s = home;
    s += "/.ce_vars";
    ifstream fp(s.c_str());
    while(fp.good())
    {
        getline(fp, s);
        if(!fp.good())break;
        if(!s.size())continue;
        if(s[0]=='#')continue;
        p.Parse(s, true);
    }
}

static void SaveVariable(const string &var, const string &value)
{
    const char *home = getenv("HOME");
    string infn = home;
    infn += "/.ce_vars";
    string tmpfn = tmpnam(NULL);
    string s;
    
    istream *fp;
    ostream *fo;
    
    fp = new ifstream(infn.c_str());
    fo = new ofstream(tmpfn.c_str());
    while(fp->good())
    {
        getline(*fp, s);
        if(!fp->good())break;
        if(!s.size() || s[0]=='#')
        {
            *fo << s << endl;
        }
        else
        {
            string tvar = p.setvarname(s, false);
            if(tvar != var)
            	*fo << tvar << " = " << s << endl;
        }
    }
    if(value != "0#" && value != "0")
		*fo << var << " = " << value << endl;

	*fo << "ans = " << value << endl;
	
    delete fp;
    delete fo;
    
    fp = new ifstream(tmpfn.c_str());
    fo = new ofstream(infn.c_str());
    while(fp->good())
    {
        getline(*fp, s);
        if(!fp->good())break;
        *fo << s << endl;
    }
    
    delete fp;
    delete fo;
    
    remove(tmpfn.c_str());    
}

string Parser::setvarname(string &s, bool checker) return resultvar
{
    if(s.size() && isidchar(s[0]) && !isdigit(s[0]))
    {
        unsigned a = 0;
        do { resultvar += s[a++]; } while(a < s.size() && isidchar(s[a]));
        unsigned b = a;
        while(a < s.size() && isspace(s[a]))++a;
        if(a < s.size() && s[a] == '=')
        {
            if(checker && Yks.Test(resultvar) == resultvar.size())
            {
                s = "";
                throw BadInput("Attempt to redefine a built-in unit");
            }
            else if(checker && Vak.Test(resultvar) == resultvar.size())
            {
                s = "";
                throw BadInput("Attempt to redefine the built-in constant '" + resultvar + "'");
            }
            else
            {
                resultvar = s.substr(0, b);
                s.erase(0, a+1);
                delblank(s);
            }
        }
        else
        	resultvar = "";
    } 
}

string Parser::DumpMerkNum(merknum e) return retval
{
    if(e.value() < 0)
    {
        retval += '-';
        e.negate();
    }
    
    const unsigned precision = e.precision();
    
    // cout << "Analysoidaan " << e.value() << " (" << precision << ")\n";
    if(e.value() == 0)
    {
	    if(precision==1) return "0";
    	if(precision==TARKKUUS) return "0#";
    }	
    
    /*
    Ekaksi kerroin sovitetaan sellaiseksi, ett luku on >= 1, mutta <= pow10(tarkkuus).
    Kokonaislukuosan desimaalien mr on log10(luku) (pyristettyn ylspin).
    Tmn jlkeen selvitetn tarvittavien desimaalien mr, ja pyristetn.
    Lopuksi tulostetaan tarvittaessa *10^jotain, mikli kerrottiin jollain.
    */
    
    // Se on nyt sitten positiivinen.
    real v = e.value();
    int kerroin = 0;
    if(v < 1)
    {
        /* Ei tykt 0.00000001 -luvuista. 0.001 on viel siedettv. */
        if(VirallinenTyyli || v < 0.001)
        {
            // Pyrist negatiivisen luvun alaspin
            kerroin = (int)floorl(log10l(v));
            //cout << endl << v << "/10^" << kerroin;
            v /= pow10l(kerroin);
            //cout << "=" << v << endl;
        }
    }
    else
    {
        real rajaarvo = pow10l(precision);
        const int kokotarkkuus = (int)floorl(log10l(v)) + 1;
        const int desimtarkkuus = precision - kokotarkkuus;
        if(v + 9*pow10l(desimtarkkuus-1) >= rajaarvo)
        {
            kerroin = (int)log10l(v / rajaarvo) + 1;
            v /= pow10l(kerroin);
        }
    }
    
    int desimadd = 0;
    
    if(!VirallinenTyyli)
    {
        /*  Esim 500.14: tarkkuus = 2. Normaali tulos = 50E1;
            v:n kokonaislukuosa olisi tss vaiheessa 50, desimaaliosa 0.14, kerroin 1 (10^1).
        */
        // Hvitetn ylimriset desimaalit.
        {
            const int kokotarkkuus = (int)floorl(log10l(v)) + 1;
            const int desimtarkkuus = precision - kokotarkkuus;
            /*
            cout << "\nBefore: " << v << " = "
                 << kokotarkkuus << '.' << desimtarkkuus 
                 << " of " << precision << endl;
            */
            const real powi = pow10l(desimtarkkuus);
            v *= powi;
            v += 1E-13L;
            modfl(v, &v);
            v /= powi;
            
            //cout << "After: " << v << endl;
            // v -= fmodl(v, pow10l(-desimtarkkuus));
        }
        // Seuraavaksi pit selvitt, montako nollaa tss on.
        unsigned nollamaara;
        // Selvitmme sen loopilla. TODO: Keksi parempi menetelm.
        for(nollamaara=0; nollamaara<precision; ++nollamaara)
            if(fmodl(v, pow10l(nollamaara+1)) != 0)
                break;
        //cout << "Siin siis on " << nollamaara << " nollaa, luvussa " << v << " siis.\n";
        unsigned merkitsevia = (unsigned)log10l(v) + 1 - nollamaara; // miinus nollien mr. 
        //cout << "Merkitsevi: " << merkitsevia << endl;
        /*
         Nyt siis jos luku on 500, 3 - 2 = 1.
         Jos merkitsevi on enemmn kuin niit pitisi olla...
         Ok, se on mahdotonta. Luku on jaettu jo tarpeeksi pieneksi.
         Jos merkitsevi on liian vhn, on ongelma ksill.
         Tm on siis esimerkiksi, jos luku on 500, ja halutaan 3 merkitsev.
         Tllin on pakko ilmoittaa 5.00E2.
        */
        if(merkitsevia < precision)
        {
            unsigned ero = precision - merkitsevia;
            if(ero <= nollamaara)
            {
                kerroin += ero;
                v /= pow10l(ero);
                // cout << "ero = " << ero << ", uusi v on nyt " << v << endl;
            }
        }
        else if(merkitsevia == precision)
        {
            // Jos luku on 500, ja halutaan 1 merkitsev.
            // Virallinen tyyli tuotta tuosta 5E2.
            // Koska pilkkua edeltvt nollat eivt ole nyt merkitsevi,
            // mys "500" on sallittu muoto.
            if(kerroin > 0)
            {
                v *= pow10l(kerroin);
                desimadd = kerroin;
                kerroin = 0;
            }
        }
    }
    const int kokotarkkuus = (int)floorl(log10l(v)) + 1;
    const int desimtarkkuus = precision - kokotarkkuus + desimadd;
    
    char Buf[256];
    sprintf(Buf, "%.*Lf", desimtarkkuus, v);
    retval += Buf;
    if(kerroin != 0)
    {
        sprintf(Buf, "E%d", kerroin);
        retval += Buf;
    }
    //cout << '(' << kokotarkkuus << " of " << precision << ", desimtarkkuus=" << desimtarkkuus << ')';
}

int Parser::SuggestKerroin(const merknum &e, bool SentitYms = false)
{
    real tmp = e.value();
    int ker = 0;
    
    // Tm on tarkempaa kuin 10-jakolooppi
    ker = (int)floorl(log10l(tmp));
    
    if(!SentitYms || ker < -6 || ker > 0)
        if(ker < 0)
            ker = int((ker-2)/3)*3;
        else
            ker = int(ker/3)*3;
    
    if(ker > 18)ker = 18;
    else if(ker < -18)ker = -18;
    
    return ker;
}

string Parser::DumpResult(elemp result, int suggkerr = 0) return retval
{
    if(dumpdebug && (result->type=='+' || result->type=='*'))
    {
        retval += "\33[1;32m";
        retval += result->type;
        retval += "\33[m";
    }
    
    switch(result->type)
    {
        case '+':
        {
            summa &s = *(summa *)&*result;
            
            bool iniplus = false;
            
            if(s.lisays != 0 || s.elems.size() == 0)
            {
                retval += DumpMerkNum(s.lisays);
                iniplus = true;
            }
            
            for(summa::elemap::const_iterator i = s.elems.begin(); i != s.elems.end(); ++i)
            {
                merknum tmp = i->second;
                bool si = false;
                bool cent = false;
                switch(i->first->type)
                {
                    case '*':
                    {
                        tulo &s = *(tulo *)&*(i->first);
                        tulo::elemap::const_iterator j;
                        bool eka = true;
                        for(j = s.elems.begin(); j != s.elems.end(); ++j)
                        {
                            if(Grammahallinta && j->first->type=='g')
                                tmp /= powl(1000, j->second);
                            else if(j->first->type!='+' && j->first->type!='*' && eka)
                            {
                                Yksikot::si_case s = Yks.SI(j->first->type);
                                if(s != Yksikot::no_si && SIhallinta)
                                {
                                    if(s == Yksikot::si_c)cent = true;
                                    si = true;
                                }
                            }
                            eka = false;
                        }
                        break;
                    }
                    case '+':
                        break;
                    default:
                        if(Grammahallinta && i->first->type=='g')
                            tmp /= 1000;
                        else
                        {
                            Yksikot::si_case s = Yks.SI(i->first->type);
                            if(s != Yksikot::no_si && SIhallinta)
                            {
                                if(s == Yksikot::si_c)cent = true;
                                si = true;
                            }
                        }
                }
                
                int suggkerr = 0;                    
                if(si)
                {
                    suggkerr = SuggestKerroin(tmp, cent);
                    tmp /= pow10l(suggkerr);
                }
                if(tmp >= 0)
                {
                    if(iniplus)retval += " + ";
                    retval += DumpMerkNum(tmp);
                }
                else
                {
                    retval += " - ";
                    retval += DumpMerkNum(-tmp);
                }
                retval += ' ';
                retval += DumpResult(i->first, suggkerr);
                iniplus = true;
            }
            break;
        }
        case '*':
        {
            tulo &s = *(tulo *)&*result;
            
            tulo::elemap::const_iterator i;
            
            bool eka=true;
            
            for(i = s.elems.begin(); i != s.elems.end(); ++i)
            {
                if(i->second < 0)continue;
                bool plus = i->first->type=='+' /*&& ((summa *)&*i->first)->size()*/;
                if(plus)retval += '(';
                if(suggkerr && Yks.SI(i->first->type) != Yksikot::no_si)
                {
                    retval += Kert.Find(suggkerr);
                    //cout << '(' << suggkerr << ')';
                    suggkerr = 0;
                }
                retval += DumpResult(i->first);
                if(plus)retval += ')';
                if(i->second == 2)retval += '';
                else if(i->second == 3)retval += '';
                else if(i->second != 1){ char Buf[64]; sprintf(Buf, "^%Lg", (real)i->second); retval += Buf; }
                eka = false;
            }
            for(i = s.elems.begin(); i != s.elems.end(); ++i)
            {
                if(i->second >= 0)continue;
                if(eka) { /*retval += '1'; */eka = false; }
                retval += '/';
                bool plus = i->first->type=='+' /*&& ((summa *)&*i->first)->size()*/;
                if(plus)retval += '(';
                if(suggkerr && Yks.SI(i->first->type) != Yksikot::no_si)
                {
                    retval += Kert.Find(suggkerr);
                    //cout << '(' << suggkerr << ')';
                    suggkerr = 0;
                }
                retval += DumpResult(i->first);
                if(plus)retval += ')';
                if(i->second == -2)retval += '';
                else if(i->second == -3)retval += '';
                else if(i->second != -1){ char Buf[64]; sprintf(Buf, "^%Lg", (real)i->second); retval += Buf; }
            }
            
            if(!s.size())retval += '1';
            
            break;
        }
        
        default:
            if(suggkerr)
            {
                retval += Kert.Find(suggkerr);
                //cout << '(' << suggkerr << ')';
                suggkerr = 0;
            }
            if(Grammahallinta && result->type=='g')retval += 'k'; // Grammat aina kilogrammoiksi
            retval += Yks[result->type];
    }
}

void Dump(elemp res)
{
    cout << p.DumpResult(res) << endl;
}

static void SegTrap(int sig)
{
    cout << "Error: Unexpected input\n";
    raise(sig);
}

#define VERSIONSTR \
    "ce "VERSION" copyright (C) 1992,2000 Bisqwit (http://iki.fi/bisqwit/)\n" \
    "This program is under GPL. ce-"VERSION".{rar,zip,tar.{gz,bz2}}\n" \
    "are available at the homepage of the author.\n"

#include "args.hh"
class Handle : public arghandler
{
    bool Help;
    string opt_h(const string &s)
    {
        Help = true;
        return s;
    }
    string opt_v(const string &s)
    {
        dumpdebug = true;
        return s;
    }
    string opt_V(const string &)
    {
        cout << VERSIONSTR;
        exit(0);
    }
    #define bool_opt(yes, no, var) string yes(const string &s) { var = true; return s; } \
                                   string no(const string &s) { var = false; return s; }
    bool_opt(opt_m1, opt_m0, VirallinenTyyli)
    bool_opt(opt_s1, opt_s0, SummaDebug)
    bool_opt(opt_t1, opt_t0, TuloDebug)
    bool_opt(opt_c1, opt_c0, CombineDebug)
    bool_opt(opt_O1, opt_O0, Optimization)
    bool_opt(opt_kg, opt_nkg,Grammahallinta)
    bool_opt(opt_si, opt_nsi,SIhallinta)
    
    #undef bool_opt
public:
    string expression;
public:
    Handle(const char *defopts, int argc, const char *const *argv)
    : arghandler(defopts, argc, argv),
        Help(false)
    {
        add("-h",  "--help",        "This text", &(Handle::opt_h));
        add("-m0", "--no-stdsignif","500 contains one significant digit (default)", &(Handle::opt_m0));
        add("-m1", "--stdsignif",   "500 contains three significant digits", &(Handle::opt_m1));
        add("-O0", "--no-optimization", "Supresses the formula expansion", &(Handle::opt_O0));
        add("-O1", "--optimization",  "Optimizes the formula (default)", &(Handle::opt_O1));
        add("-s0", "--no-summadebug", "(default)", &(Handle::opt_s0));
        add("-s1", "--summadebug",   "Debygs sum operations", &(Handle::opt_s1));
        add("-t0", "--no-tulodebug", "(default)", &(Handle::opt_t0));
        add("-t1", "--tulodebug",   "Debygs product operations", &(Handle::opt_t1));
        add("-c0", "--no-combinedebug", "(default)", &(Handle::opt_c0));
        add("-c1", "--combinedebug",   "Debygs combining operations", &(Handle::opt_c1));
        add("-V",  "--version",     "Version info", &(Handle::opt_V));
        add("-v",  "--verbose",     "Another debugging option", &(Handle::opt_v));
        add("-nkg","--no-grammahallinta", "Don't favor kg", &(Handle::opt_nkg));
        add("-kg", "--grammahallinta", "Favor kg (default)", &(Handle::opt_kg));
        add("-nsi","--no-si-control", "Don't use m,k,M,c etc", &(Handle::opt_nsi));
        add("-si", "--si-control", "Use SI-prefices (default)", &(Handle::opt_si));
        parse();
    }
    virtual void defarg(const string &s)
    {
        // cout << "defarg(\"" << s << "\")\n";
        if(s.size())
        {
            if(expression.size())
                expression += ' ';
            expression += s;
        }
    }
    virtual void parse()
    {
        arghandler::parse();
        if(Help)
        {
            cout << "Usage: " << a0 << " [options] <expression>\n"
                    "\nOptions:\n";
            
            listoptions();
            
            cout << endl;
            
            exit(0);
        }
    }
};

int main(int argc, const char *const *argv)
{
    signal(SIGSEGV, SegTrap);
    
    Handle parameters("", argc, argv);
    string s = parameters.expression;
    
    if(!s.size())
    {
        cout << "Anna lauseke: " << flush;
        getline(cin, s);
    }
    
    // cout << "Parsing: '" << s << "'\n";
        
    p.Parse(s);
    
    return 0;
}
